/* these functions read and write a raw unsigned char buffer[] from a
 * read()  call to a CAD4_Buffer data type (see cad4io.h).
 */

#define _XOPEN_SOURCE
#include <string.h>
#include <stdio.h>
#include "cad4io.h"


/* initializes (zeros) a CAD4_Buffer struct */

void init_cad4_buffer( CAD4_Buffer *b )
{
    b->header_byte = 0;
    b->length.data_len = 0;
    memset( &(b->data.c), '\0', IO_DATA_LEN + IO_CRC_LEN );

    return;
} 

unsigned char get_error_flag( CAD4_Buffer *buf )
{
    unsigned char ch;

    ch = buf->header_byte & M_EFL;
    return  ch;
}

unsigned char get_sequence_number( CAD4_Buffer *buf )
{
    unsigned char ch;

    ch = buf->header_byte & M_SEQ;
    return  ch;
}

unsigned char get_function_code( CAD4_Buffer *buf )
{
    unsigned char ch;

    ch = buf->header_byte & M_FUN;
    return  ch;
}

/* this next function was essentially written by Chris Torek 
 * (torek@elf.bsdi.com) in response to my plea for help on 
 * comp.lang.c.  w0 contains the low 16 bits, the w1 word contains
 * the high 15 bits (as it's low 15 bits) and a sign bit (as it's
 * bit 15.
 */

double conv_hw_words( hw_word w1, hw_word w0 )
{
    long   v31;   /* 31 bit value from hardware */
    double v32;   /* 32 bit value including sign */

    if( w1 & 0x8000 ) {
        v31 = ((( ~w1 & 0x7fff) << 16) | ( ~w0 & 0xffff ) );
        v32 = -v31 - 1.0; /* the -1.0 here is because negative values
                           * are twos complement, so that (ffff, ffff)
                           * represents -1 and not -0
                           */
    }
    else {
        v31 = (( w1 & 0x7fff) << 16 ) | w0;
        v32 = v31;
    } 

    return v32;
}

/* fill_cad4_data_block() -- takes a CAD4_Buffer and fills a
 * cad4_data_block struct with the information contained in the
 * the buffer.
 */
void fill_cad4_data_block( CAD4_Buffer *buf, struct cad4_data_block *blk )
{
    blk->nswreg = buf->data.s[C4H_SWREG];
    blk->iroutf = buf->data.s[C4H_ROUTFL];
    blk->incr_1 = buf->data.s[C4H_INCR];
    blk->incr_2 =  0;
    blk->npi_1  =  0;
    blk->npi_2  =  0;
    blk->iscanw =  0;
    blk->motw   =  0;
    blk->ishutf =  0;
    blk->ibalf  =  0;
    blk->iattf  =  0;
    blk->iresf  =  0;
    blk->ierrf  = buf->data.s[C4H_ERRFL];
    blk->intfl  = buf->data.s[C4H_INTFL];

/* the quantities below are floats and need special conversions */

    blk->xrayt  =       (float)conv_hw_words( buf->data.s[C4H_XRTIM],
                                              buf->data.s[C4H_XRTIM+1] );
    blk->tthp   =       (float)conv_hw_words( buf->data.s[C4H_TTHMXH],
                                              buf->data.s[C4H_TTHMXH+1] ) /
                                              ( STDGR * -32.0 );
    blk->tthn   =       (float)conv_hw_words( buf->data.s[C4H_TTHMNH],
                                              buf->data.s[C4H_TTHMNH+1] ) /
                                              ( STDGR * -32.0 );

    blk->aptw   =       (float)conv_hw_words( buf->data.s[C4H_APWH],
                                              buf->data.s[C4H_APWH+1] );
    blk->want[FOR_PH] = (float)conv_hw_words( buf->data.s[C4H_PHWH],
                                              buf->data.s[C4H_PHWH+1] ) / STDGR;
    blk->want[FOR_OM] = (float)conv_hw_words( buf->data.s[C4H_OMWH],
                                              buf->data.s[C4H_OMWH+1] ) / STDGR;
    blk->want[FOR_KA] = (float)conv_hw_words( buf->data.s[C4H_KAWH],
                                              buf->data.s[C4H_KAWH+1] ) / STDGR;
    blk->want[FOR_TH] = (float)conv_hw_words( buf->data.s[C4H_THWH],
                                              buf->data.s[C4H_THWH+1] ) / STDGR;
    blk->aptm         = (float)conv_hw_words( buf->data.s[C4H_APMH],
                                              buf->data.s[C4H_APMH+1] ) / STDGR;
    blk->cmeas[FOR_PH] = (float)conv_hw_words( buf->data.s[C4H_PHMH],
                                              buf->data.s[C4H_PHMH+1] ) / STDGR;
    blk->cmeas[FOR_OM] = (float)conv_hw_words( buf->data.s[C4H_OMMH],
                                              buf->data.s[C4H_OMMH+1] ) / STDGR;
    blk->cmeas[FOR_KA] = (float)conv_hw_words( buf->data.s[C4H_KAMH],
                                              buf->data.s[C4H_KAMH+1] ) / STDGR;
    blk->cmeas[FOR_TH] = (float)conv_hw_words( buf->data.s[C4H_THMH],
                                              buf->data.s[C4H_THMH+1] ) / STDGR;
    return;

}

void print_cad4_data_block( FILE *out, struct cad4_data_block *blk )
{
    fprintf( out, "Switch Reg:\t%ho\t\t\tMotor Selection:\t%#hx\n"
                  "Routine Flag:\t%#hx\t\t\tShutter Flag:\t\t%#hx\n"
                  "Master Incr:\t%#hx\t\t\tBalanced Filter Flag:\t%#hx\n"
                  "Slave Incr:\t%#hx\t\t\tAttenuator Flag:\t%#hx\n"
                  "Master Speed:\t%#hx\t\t\tReserve Flag:\t%#hx\n"
                  "Slave Speed:\t%#hx\t\t\tResult Err Flag:\t%#hx\n"
                  "Scan Width:\t%#hx\t\t\tIntensity Result Flag:\t%#hx\n", 
                  blk->nswreg, blk->motw, 
                  blk->iroutf, blk->ishutf,
                  blk->incr_1, blk->ibalf,
                  blk->incr_2, blk->iattf,
                  blk->npi_1, blk->iresf,
                  blk->npi_2, blk->ierrf, 
                  blk->iscanw, blk->intfl );

   fprintf( out, "Theta wanted:\t%8.3f\t\tTheta measured:\t%8.3f\n"
                 "Phi wanted:\t%8.3f\t\tPhi measured:\t%8.3f\n"
                 "Omega wanted:\t%8.3f\t\tOmega measured:\t%8.3f\n"
                 "Kappa wanted:\t%8.3f\t\tKappa measured:\t%8.3f\n"
                 "Apt wanted:\t%4.2f\t\t\tApt measured:\t%4.2f\n"
                 "X-ray Exposure Time:\t%10.1f\n"
                 "Theta max:%8.3f\t\tTheta min: %8.3f\n\n", 
                 blk->want[FOR_TH], blk->cmeas[FOR_TH],
                 blk->want[FOR_PH], blk->cmeas[FOR_PH],
                 blk->want[FOR_OM], blk->cmeas[FOR_OM],
                 blk->want[FOR_KA], blk->cmeas[FOR_KA],
                 blk->aptw, blk->aptm, 
                 blk->xrayt, 
                 blk->tthp, blk->tthn );
    return;
}
