/* C header file containing I/O information for DIFRAC (written by
 * Peter White and Eric Gabe.  This is a start at my attempt at a C language
 * implementation of the low level serial port communication and is the
 * rough equivalent of the FORTRAN CAD4COMM include file from DIFRAC.
 * Here is the CAD4 buffer format used by DIFRAC:
 * Cad4 buffer format
 *
 *       15                                              0
 *       +-----------------------+
 *       +   header byte         +
 *       +------------------------------------------------+
 *       +  length(high byte)    +   length (low byte)    +
 *       +------------------------------------------------+
 *       !      switch register word or load address      +
 *       +------------------------------------------------+
 *       !      switch register word or load address      +
 *       +------------------------------------------------+
 *       !                                                !
 *       !                       ...                      !
 *
 *       !               256 words data (512. bytes)      !
 *       +------------------------------------------------+
 *       +               CRC (16. bit)                    +
 *       +------------------------------------------------+
 *
 *      Implemented by:
 *	Paul D. Boyle
 *	Department of Chemistry
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *
 *	boyle@laue.chem.ncsu.edu
 */
#ifndef CAD4IO_H
#define CAD4IO_H
/* QIO status codes */

#define  SS_NORMAL    ((int) 0x1)
#define  SS_TIMEOUT   ((int) 0x22c)

/* Internal status codes returned in one byte */
/* May 1999: these were originally cast as (unsigned short), but we never use 
 * the higher 8 bits, so we cast them to (unsigned char)
 */

#define   E_SUC     ((unsigned char) 0x0)     /* Success */
#define   E_TOS     ((unsigned char) 0x4)     /* tmo + data */
#define   E_TOL     ((unsigned char) 0x8)     /* tmo + no data */
#define   E_SEQ     ((unsigned char) 0xc)     /* Unexpected sequence number */
#define   E_CRC     ((unsigned char) 0x10)    /* CRC error */
#define   E_TYP     ((unsigned char) 0x14)    /* Unexpected function code */
#define   E_OVF     ((unsigned char) 0x18)    /* Buffer overflow */
#define   E_PND     ((unsigned char) 0x1c)    /* Any system service fail */

/* Masks of various sorts */

#define   M_SEQ     ((unsigned char) 0x3)     /* Mask to get sequence bits */
#define   M_EFL     ((unsigned char) 0x1c)    /* Mask to get error flag */
#define   M_FUN     ((unsigned char) 0xe0)    /* Mask to get function bits */

/* Function codes for CAD4 IO routine (were originally unsigned shorts) */

#define   FN_INIT        ((unsigned char) 0x0)      /* Bootstrap 11/02 */ 
#define   FN_XFR_ASC     ((unsigned char) 0x20)     /* Xfr ASCII buffer to 11/02 */
#define   FN_XFR_MEM     ((unsigned char) 0x40)     /* Xfr code block to 11/02 */
#define   FN_TR_SWR      ((unsigned char) 0x60)     /* trm and rec. SWR */

/* these last 4 defines were originally unsigned shorts of the form 
 * 0xff80, 0xffa0, 0xffc0, 0xffe0
 * 'host' is the term used here to refer to the computer running
 * the diffractometer control software. e.g. a PC
 */
#define   FN_TR_GON      ((unsigned char) 0x80)   /* Trm. and rec. goniometer data */
#define   FN_TR_ASC      ((unsigned char) 0xa0)   /* trm and rec. ASCII buffer */ 
#define   FN_REQ_MEM     ((unsigned char) 0xc0)   /* req. code block from host  */
#define   FN_REQ_ASC     ((unsigned char) 0xe0)   /* req. ASCII buffer from host */

/* Define Baud rate constants  for CAD4 terminal */

#define   BAUD_38400     ((unsigned short) 3)
#define   BAUD_19200     ((unsigned short) 6)
#define   BAUD_9600      ((unsigned short) 12)
#define   BAUD_4800      ((unsigned short) 24)
#define   BAUD_2400      ((unsigned short) 48)
#define   BAUD_1200      ((unsigned short) 96)
#define   BAUD_600       ((unsigned short) 192)
#define   BAUD_300       ((unsigned short) 384)

/* Define default system constants as will be expected at bottom
 * of LSI target computer.
 */

#define   LSI_BOTTOM     ((unsigned short) 0x8000)
#define   SBC_BOTTOM     ((unsigned short) 0xec00)

#define   SBCP_BOTTOM    ((unsigned short) 0x8000)

/* Defines for DVI system service */

#define  DVI_DEVCLASS    ((int) 0x00000004)
#define  DVI_DEVDEPEND   ((int) 0x0000000a)
#define  DVI_DEVDEPEND2  ((int) 0x0000001c)
#define  DVI_DEVTYPE     ((int) 0x00000006)

/* Define word for CRC */

#define  ICONST          ((int) 0xa001)

/* Define for converting degrees to encoder values and 
 * vice-versa
 */

#define  DGRST (360.0 /(128.0 * 4096.0))  /* encoder -> degrees */
#define  STDGR (128.0 * 4096.0 / 360.0)   /* degrees -> encoder */

/* suggested by Chris Torek (torek@elf.bsdi.com) from correspondance
 * based on my comp.lang.c posting -- typedef a type for a "hardware
 * word".  He suggests an unsigned int.  We go with that suggestion.
 */

typedef  unsigned int hw_word;

/* Define CAD4 handler offsets these have been adjusted due to
 * difference between FORTRAN and C array offsets.  To get back
 * just + 1 to the values except for the ones marked 'original value'.
 */

#define  C4H_SWREG       ((unsigned short)  0)
#define  C4H_ROUTFL      ((unsigned short)  1)
#define  C4H_ERRFL       ((unsigned short)  2)
#define  C4H_INTFL       ((unsigned short)  3)
#define  C4H_TTHMXH      ((unsigned short)  4)
#define  C4H_TTHMNH      ((unsigned short)  6)
#define  C4H_SASYSC      ((unsigned short)  8)
#define  C4H_XRTIM       ((unsigned short)  9)
#define  C4H_MSELW       ((unsigned short)  11)
#define  C4H_NRD         ((unsigned short)  12)
#define  C4H_NID         ((unsigned short)  13)
#define  C4H_INCR        ((unsigned short)  11)
#define  C4H_INCI        ((unsigned short)  0)  /* original value */
#define  C4H_DINCR       ((unsigned short)  1)  /* original value */
#define  C4H_NRINC       ((unsigned short)  2)  /* original value */
#define  C4H_THWH        ((unsigned short)  29)
#define  C4H_PHWH        ((unsigned short)  31)
#define  C4H_OMWH        ((unsigned short)  33)
#define  C4H_KAWH        ((unsigned short)  35)
#define  C4H_APWH        ((unsigned short)  37)
#define  C4H_APWL        ((unsigned short)  38)
#define  C4H_THMH        ((unsigned short)  39)
#define  C4H_PHMH        ((unsigned short)  41)
#define  C4H_OMMH        ((unsigned short)  43)
#define  C4H_KAMH        ((unsigned short)  45)
#define  C4H_APMH        ((unsigned short)  47)
#define  C4H_DUMP_0      ((unsigned short)  49)  /* That's a 0 (zero) not an 'o' */

/* Defines for the c4h_routfl table */

#define  RF_SWI          ((unsigned short)  0x0)
#define  RF_MEA          ((unsigned short)  0x4)
#define  RF_COL          ((unsigned short)  0x8)
#define  RF_POC          ((unsigned short)  0x10)
#define  RF_POS          ((unsigned short)  0x20)
#define  RF_POF          ((unsigned short)  0x40)
#define  RF_SAP          ((unsigned short)  0x80)
#define  RF_SCA          ((unsigned short)  0x100)
#define  RF_SCD          ((unsigned short)  0x200)
#define  RF_RES          ((unsigned short)  0x8000)

#define ROUT_0           ((unsigned short) RF_SWI + RF_RES)
#define ROUT_1           ((unsigned short) RF_SWI + RF_MEA + RF_RES)
#define ROUT_2           ((unsigned short) RF_SWI + RF_COL + RF_RES)
#define ROUT_3           ((unsigned short) RF_SWI + RF_POS + RF_RES)
#define ROUT_4           ((unsigned short) RF_SWI + RF_POF + RF_RES)
#define ROUT_5           ((unsigned short) RF_SWI + RF_POC + RF_POF + RF_RES)
#define ROUT_6           ((unsigned short) RF_SWI + RF_SCA + RF_RES)
#define ROUT_7           ((unsigned short) RF_SWI + RF_SAP + RF_SCA + RF_RES)
#define ROUT_8           ((unsigned short) RF_SWI + RF_POC + RF_POF + RF_SCA + RF_RES)
#define ROUT_9           ((unsigned short) RF_SWI + RF_SCD + RF_RES)
#define ROUT_10          ((unsigned short) RF_SWI + RF_SAP + RF_SCD + RF_RES)
#define ROUT_11          ((unsigned short) RF_SWI + RF_POC + RF_POF + RF_SAP + RF_SCD + RF_RES)
#define ROUT_12          ((unsigned short) RF_SWI + RF_POC + RF_RES)
#define ROUT_13          ((unsigned short) RF_SWI + RF_RES) /* Free */
#define ROUT_14          ((unsigned short) RF_SWI + RF_RES) /* Free */
#define ROUT_15          ((unsigned short) RF_SWI + RF_RES) /* Free */

/* Add those errtbl and inttbl values later */

/* CAD4 handler sasync table */

#define  SASYNC_ATT      ((unsigned short)  0x2000)
#define  SASYNC_SHU      ((unsigned short)  0x8000)

#define  SAS_0           ((unsigned short)  0x0)
#define  SAS_1           ((unsigned short)  0x2000)  /* same as SASYNC_ATT */
#define  SAS_2           ((unsigned short)  0x8000)  /* same as SASYNC_SHU */
#define  SAS_3           ((unsigned short)  SASYNC_ATT + SASYNC_SHU) 

/* Defines for angles  these are different in C because arrays are
 * read from element 0 rather than 1 as in Fortran.
 */

#define  FOR_PH          ((unsigned short)  0)
#define  FOR_OM          ((unsigned short)  1)
#define  FOR_KA          ((unsigned short)  2)
#define  FOR_TH          ((unsigned short)  3) 

/* Motor table defines */

#define  NO              ((unsigned short) 0)
#define  AP              ((unsigned short) 1)
#define  PH              ((unsigned short) 2)
#define  OM              ((unsigned short) 3)
#define  KA              ((unsigned short) 4)
#define  TH              ((unsigned short) 5)

/* Structures and typedef's for the cad4_communication routines */

struct sys_par {
       unsigned short hivolt;
       unsigned short lower_level;
       unsigned short window;
       unsigned short syspar_def_4;
       unsigned short dead_time;
       unsigned short term_baud;
       unsigned short default_baud;
       unsigned short sypar_def_8;
       unsigned short system_clock;
       unsigned short pos_step;
       unsigned short theta_mot_gain;
       unsigned short phi_mot_gain;
       unsigned short omega_mot_gain;
       unsigned short kappa_mot_gain;
       unsigned short dial_mot_gain;
       unsigned short system_flag; 
       unsigned short max_emission;
       };

struct cad4_data_block {
       unsigned short nswreg;     /* Slave switch register */
       unsigned short iroutf;     /* Routine flag */
       unsigned short incr_1;     /* Master increment */
       unsigned short incr_2;     /* Slave increment */
       unsigned short npi_1;      /* Inverse scan speed for master */
       unsigned short npi_2;      /* Relative scan speed for slave */
       unsigned short iscanw;     /* scan width */
       unsigned short motw;       /* Motor selection word */
       unsigned short ishutf;     /* shutter flag */
       unsigned short ibalf;      /* balanced filter flag */
       unsigned short iattf;      /* attenuator filter flag */
       unsigned short iresf;      /* reserve flag */
       unsigned short ierrf;      /* result error flag */
       unsigned short intfl;      /* Intensity result flag */
       float xrayt;      /* X-ray time */
       float tthp;       /* limit value for theta (pos.) */
       float tthn;       /* limit value for theta (neg.) */
       float aptw;       /* wanted encoder value for apeture */
       float want[4];    /* wanted values for gonio-angles */
       float spare[6];   /* spare locs */
       float aptm;       /* measured value of apeture */
       float cmeas[4];   /* measured gonio-angles */
       float dump[512];  /* intensity dumps */
       };

struct cad4_ca_comm {
       unsigned short ibycan[2];
       unsigned short ir5can;
       unsigned short lsypar;     /* LSI syspar address */
       unsigned short io_coswr;   /* switch options register from host-> 11/02 */
       unsigned short io_cobnr;   /* Number of calls to 11/02 */
       char io_cohex;    /* header from host -> 11/02 */
       };

/* build up structures for the communication buffer 
 */

#define IO_DATA_LEN  518  /* this includes the appended crc for the data */
#define IO_HEADER_LEN  3
#define IO_CRC_LEN     (sizeof(unsigned short))

#define HEADER_BYTE    0
#define LOW_BYTE       1
#define HIGH_BYTE      2
#define START_DATA     3

/* the following unions are not portable, and will not work for
 * CPU which have a different endianess than the LSI interface.
 */
union cad4_data {
      unsigned char  c[IO_DATA_LEN];
      unsigned short s[IO_DATA_LEN/2];
      };

union data_len {
      unsigned char ch[2];
      unsigned short data_len;
      };

struct cad4_buffer {
       unsigned char   header_byte;
       union data_len  length;
       union cad4_data data;
       };
typedef struct cad4_buffer CAD4_Buffer;

/* Function Prototypes */

/* these first 2 function prototypes are strictly P. Boyle's 
 * work.  Utility functions to manage the CAD4_Buffer to serial
 * port data transfers.
 */
void init_cad4_buffer( CAD4_Buffer *buf );
/*
void data_to_buffer( CAD4_Buffer *dst, unsigned char src[], size_t n_bytes );
void buffer_to_data( unsigned char *dst, CAD4_Buffer src, size_t n_bytes );
*/
unsigned char get_error_flag( CAD4_Buffer *buf );
unsigned char get_sequence_number( CAD4_Buffer *buf );
unsigned char get_function_code( CAD4_Buffer *buf );
void load_cad4_data_block( CAD4_Buffer *buf, struct cad4_data_block *blk );

/* these next prototypes should probably go somewhere else
 * eventually.
 */
float convert_shorts_to_float( unsigned short *s );
void reverse_bytes( void *f, size_t size );
void reverse_words( unsigned short *s );

#if 0  /* omit for  now */

unsigned short cad4_io( unsigned short cmd,  unsigned short (*io_pre()),
               unsigned short (*io_post0()), unsigned short (*io_post1()),
               unsigned short (*io_post2()), unsigned short (*io_post3()),
               unsigned short (*io_post4()), unsigned short (*io_post5()),
               unsigned short (*io_post6()), unsigned short (*io_post7()),
               CAD4_Buffer cad4_buffer);
unsigned short cad4_load_syspar
unsigned short cad4_get_instrument
unsigned short cad4_ini_terminal
unsigned short cad4_exit_handler
unsigned short cad4_reset_terminal
unsigned short cad4_readprompt
unsigned short cad4_prepare_output
unsigned short cad4_check_crc
unsigned short cad4_type_error( void );
#endif
#endif /* CAD4IO_H */
