/* cd4logger -- this program attaches to the shared memory segment
 * setup by the sniffer program.  It reads the shared memory and 
 * prints out the contents to either a file or to the screen.
 */

#define _XOPEN_SOURCE
/* unix/posix headers */
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <getopt.h>
#include <signal.h>

/* standard C headers */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

/* local headers */
#include "cd4port.h"
#include "cad4io.h"
#include "cd4port_ipc.h"

int main( int argc, char **argv )
{
    int           i_opt, 
                  verbose = 0,  /* don't print out buffer contents
                                 * by default */
                  ret_val;

    FILE          *logfile = stdout; /* default writes stuff to screen */
    char          cad4_nr[2]; 
    char          *cd4port_file_name,
                  *cd4logger_file_name;
    key_t         key;
    pid_t         cd4port_pid;
    COMM_Packet   packet;
    COMM_Packet   *shared_mem;
    int           sem_id,
                  shm_id;

    struct sigaction action; 

/* get the number of the instrument in which we are interested  */
    strncpy( cad4_nr, argv[argc-1], 1 );
    cad4_nr[1] = '\0';


/* build filenames for the various PID files, and read and write to them */
    cd4port_file_name = create_pid_file_name( CD4PORT_PID, cad4_nr );
    if( !cd4port_file_name ) {
        fputs( "Couldn't create cd4port PID file name. Exiting.\n", stderr );
        exit( EXIT_FAILURE );
    }

    cd4logger_file_name = create_pid_file_name( CD4LOGGER_PID, cad4_nr );
    if( !cd4logger_file_name ) {
        fputs( "Couldn't create cd4logger PID file name. Exiting.\n", stderr );
        exit( EXIT_FAILURE );
    }

    cd4port_pid = get_pid_file( cd4port_file_name );
    if( -1 == cd4port_pid ) {
        fputs( "Couldn't get cd4port PID. Exiting.\n", stderr );
        exit( EXIT_FAILURE );
    }

    write_pid_file( cd4logger_file_name );


/* process the command line arguments. */
    while( (i_opt = getopt(argc, argv, "lv")) != EOF ) {
       switch( i_opt ) {
          case 'l':     /* that's a lower case 'L' for logging */
             logfile = fopen( LOG_FILE_NAME, "w" );
             if( !logfile ) {
                 perror( LOG_FILE_NAME );
                 exit( EXIT_FAILURE );
             }
             break;
          case 'v':
             verbose = 1;
             break;
          default:
              fprintf( stderr,
                       "Error: %c -- unrecognized option.\n", i_opt );
             break;
       }
    }   


/* initialize the signal handling code */

    ret_val = sigemptyset( &action.sa_mask );
    if( -1 == ret_val ) {
        perror( "sigemptyset()" );
        exit( EXIT_FAILURE );
    }


    ret_val = sigaddset( &action.sa_mask, SIGUSR1 );
    if( -1 == ret_val ) {
        perror( "sigaddset()" );
        exit( EXIT_FAILURE );
    }

    action.sa_flags = 0;
    action.sa_handler = sigusr_handler;

    ret_val = sigaction( SIGUSR1, &action, NULL );
    if( -1 == ret_val ) {
        perror( "sigaction()" );
        exit( EXIT_FAILURE );
    }

/* initialize the SYSV IPC stuff.  Use the same key value for
 * both shared memory and semaphore calls.
 */
    key = CAD4_KEY + (key_t)strtol(cad4_nr, NULL, 10 );

    sem_id = semget( key, 1, MODE );
    if( -1 == sem_id ) {
        perror( "semget()" );
        exit( EXIT_FAILURE );
    }

    ret_val = sem_set( sem_id );
    if( -1 == ret_val ) {
        sem_delete( sem_id );
        exit( EXIT_FAILURE );
    }
    shm_id = shmget( key, sizeof(COMM_Packet), MODE );
    if( -1 == shm_id ) {
        sem_delete( sem_id );
        perror( "shmget()" );
        exit( EXIT_FAILURE );
    }
 
    shared_mem = shmat( shm_id, 0, 0 );
    if( ((COMM_Packet *)-1) == shared_mem ) {
        perror( "shmat()" );
        exit( EXIT_FAILURE );
    }

/* put keyboard into non-canonical mode */
    ret_val = non_canonical_kybd();
    if( -1 == ret_val ) {
      fputs( "Bringing console into non-canonical mode failed. Exiting.\n",
              stderr );
      restore_kybd();
      exit( EXIT_FAILURE );
    }

/* tell the user we are ready to go, and how to terminate program */
    fprintf( stdout, "Shared memory attached at %p\n", shared_mem );
    fputs( "Hit 'q' to quit program.\n", stdout );


/* signal cd4port */

    ret_val = kill( cd4port_pid, SIGUSR2 );
    if( -1 == ret_val ) {
        fprintf( stderr, "Couldn't send SIGUSR2 to PID %d: %s\n",
                                            cd4port_pid, strerror(errno) );
    }

    ret_val = kill( cd4port_pid, SIGHUP );
    if( -1 == ret_val ) {
        fprintf( stderr, "Couldn't send SIGHUP to PID %d: %s\n",
                                            cd4port_pid, strerror(errno) );
    }

    while( 1 ) {
      pause();
      sem_attempt_lock( sem_id );
      memcpy( &packet, shared_mem, sizeof(COMM_Packet) ); 
      sem_release( sem_id );
      print_cad4_packet( logfile, &packet, verbose );
      if( 'q' == key_hit() )
          break;
      else
          continue;
    }

/* clean up SYSV IPC and restore the keyboard to canonical 
 * mode.
 */
    restore_kybd();

    destroy_pid_file( cd4logger_file_name );

   
    ret_val = kill( cd4port_pid, SIGUSR2 );
    if( -1 == ret_val ) {
        fprintf( stderr, "Couldn't send SIGUSR2 to PID %d: %s\n",
                                            cd4port_pid, strerror(errno) );
    }

    fputs( "Detaching shared memory...\n", stdout );
    ret_val = shmdt( (char *)shared_mem );
    if( -1 == ret_val )
         perror( "shmdt()" );
    
    fputs( "Done.\n", stdout );
    exit( EXIT_SUCCESS );
}
