/* program for observing what is coming and going on a serial port.  This
 * is specifically used for looking at the serial port when running the 
 * CAD4 software under Linux.
 *
 * Here is the idea:
 *	* The CAD4 is configured read a pseudoterminal rather than from
 *	  "a real" serial port. That is, ~cad4/data0/config/rs232_port.cua0
 *	  is mv'ed to be rs232_port.ttypf or something that is unlikely to
 * 	  be in use by some other process.
 *	* This monitoring software opens the "real" serial port (i.e.
 *	  this port to which the CAD4 interface  is attached.  It also
 *	  opens the /dev/ptypf to talk to the control software.
 *	* The monitor passes data from the control software to the serial
 *	  port and vice versa and logs all transactions with a time stamp.
 *
 * The code should have endianess stuff put in, for the case that the control
 * software runs on an SGI. (not yet).
 *
 * Usage:
 *	cd4port -u<username> -p<serial port device> cad4_number
 *      Example:
 *	cd4port -ucad4 -pcua0 0
 *	Other options:  -S<speed>  POSIX speeds only
 *			-P<parity> o|O|e|E (odd or even)
 *                      -R enable non scrolling readout of byte transfers.
 *
 *	You shouldn't normally need to set these other options.
 *
 *	Where the <username> is the username under which the cad4 software 
 *	resides.  The cad4_number is the number of the cad4 instrument
 * 	you want monitored. 
 *
 * Written by:
 *	Paul D. Boyle
 *	Department of Chemistry
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *	boyle@laue.chem.ncsu.edu
 *	31 October 1997
 *
 *	Logging via SYSV shared memory added April 1999
 *
 *                          NO WARRANTY
 *
 * 1. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
 * FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
 * OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
 * PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
 * OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK
 * AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
 * PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
 * REPAIR OR CORRECTION.
 * 
 * 2. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
 * WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY
 * AND/OR REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU
 * FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING
 * BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR
 * LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM
 * TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * 
 */
#define _XOPEN_SOURCE

/* unix/posix headers */
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <signal.h>
#include <pwd.h>

/* standard C headers */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <errno.h>

/* local headers */
#include "cd4port.h"
#include "cad4io.h" 
#include "cd4port_ipc.h"

/* global variables */

pid_t pid;
char  *pid_file_name;  /* for getting the cd4logger's PID using
                        * sighup_handler() */
int logger_attached = 0;

int main( int argc, char **argv )
{
    int            i_opt,
                   i,                /* counting index */
                   s_set,            /* find select() limits */
#ifdef TX_INCMP_PKT
                   ser_ret_val,      /* return value for serial_comm() */
#endif
                   ret_val,          /* return value for select() */
                   ipc_ret_val,      /* return value for IPC calls */
                   fd_array[2];      /* [0] = serial port,
                                        [1] = pseudoterminal */
    fd_set         read_set;               /* for select() */
    unsigned int   n_seq = 0;        /* sequence number for serial port I/O */
    char           *serial_device,
                   *pseudo_master,
                   *pseudo_slave,
                   *cd4port_file_name;

    char           cad4_nr[2];
    int            parity = DEFAULT_PARITY;
    int            speed = DEFAULT_SPEED;

    struct passwd  *user_info;
    COMM_Packet    packet;
    COMM_Packet    *packet_ptr = NULL; 

/* SYSV IPC related variables */
    COMM_Packet    *shared_mem;
    key_t          key;
    int            shm_id,      /* shared memory identifier */
                   sem_id;      /* semaphore identifier */

/* signal handling variables. */
    struct sigaction action, 
                     usr2;
#ifdef TX_INCMP_PKT
    packet_ptr = &packet;
#endif

/* get the number of the instrument in which we are interested  */
    strncpy( cad4_nr, argv[argc-1], 1 );
    cad4_nr[1] = '\0';

/* build filenames for the various PID files, and read and write to them */
    cd4port_file_name = create_pid_file_name( CD4PORT_PID, cad4_nr );
    if( !cd4port_file_name ) {
        fputs( "Couldn't create cd4port PID file name. Exiting.\n", stderr );
        exit( EXIT_FAILURE );
    }

    pid_file_name = create_pid_file_name( CD4LOGGER_PID, cad4_nr );
    if( !pid_file_name ) {
        fputs( "Couldn't create cd4logger PID file name. Exiting.\n", stderr );
        exit( EXIT_FAILURE );
    }

    write_pid_file( cd4port_file_name );

/* initialize the signal handling code */

    ret_val = sigemptyset( &action.sa_mask );
    if( -1 == ret_val ) {
        perror( "sigemptyset()" );
        exit( EXIT_FAILURE );
    }


    ret_val = sigaddset( &action.sa_mask, SIGHUP );
    if( -1 == ret_val ) {
        perror( "sigaddset()" );
        exit( EXIT_FAILURE );
    }

    ret_val = sigemptyset( &usr2.sa_mask );
    if( -1 == ret_val ) {
        perror( "sigemptyset()" );
        exit( EXIT_FAILURE );
    }


    ret_val = sigaddset( &usr2.sa_mask, SIGUSR2 );
    if( -1 == ret_val ) {
        perror( "sigaddset()" );
        exit( EXIT_FAILURE );
    }

    action.sa_flags = 0;
    action.sa_handler = sighup_handler;

    usr2.sa_flags = 0;
    usr2.sa_handler = sigusr2_handler;

    ret_val = sigaction( SIGHUP, &action, NULL );
    if( -1 == ret_val ) {
        perror( "sigaction()" );
        exit( EXIT_FAILURE );
    }

    ret_val = sigaction( SIGUSR2, &usr2, NULL );
    if( -1 == ret_val ) {
        perror( "sigaction()" );
        exit( EXIT_FAILURE );
    }

/* process the command line args. */

    while( (i_opt = getopt(argc, argv, "Ru:p:P:S:")) != EOF ) {
       switch( i_opt ) {
          case 'p':     /* name of the serial port device */
             serial_device = malloc( strlen(optarg) +
                             strlen(DEVICE_PREFIX) + 1 );
             if( NULL == serial_device ) {
                 fprintf( stderr, "Memory Allocation error.%s:%d Exiting.\n", 
                          __FILE__, __LINE__ );
                 exit( EXIT_FAILURE );
             }
             sprintf( serial_device, "%s%s", DEVICE_PREFIX, optarg );
             break;
          case 'u':    /* user name under which the cad4 software
                          is installed */
             user_info = getpwnam( optarg );
             if( NULL == user_info ) {
                 fprintf( stderr, "No user by name of %s.  Exiting.\n",
                          optarg );
                 exit( EXIT_FAILURE );
             }
             break;
          case 'P':
             parity = (int)optarg[0];
             break;
          case 'S':
             speed = (int)strtol( optarg, (char **)NULL, 10);
             break; 
          default:
              fprintf( stderr,
                       "Error: %c -- unrecognized option.\n", i_opt );
             break;
       }
    }


/* open the real serial port */
    fd_array[0] = open( serial_device, O_RDWR | O_NOCTTY | O_NONBLOCK );
    if( -1 == fd_array[0] ) {
       fprintf( stderr, "Errno = %d at line %d in file %s\n",
                errno, __LINE__, __FILE__ );
       perror( serial_device );
       destroy_pid_file( cd4port_file_name ); 
       exit( EXIT_FAILURE );
    }      
    if( -1 == terminal_init(fd_array[0], speed, parity) ) {
        fprintf( stderr, "Problem initializing %s. Exiting.\n",
                 serial_device );
        destroy_pid_file( cd4port_file_name ); 
        exit( EXIT_FAILURE );
    }

    fprintf(stdout, "Serial port %s\nOpened and initialized. File descriptor %d\n", 
            serial_device, fd_array[0] );

/* get the cad4 device specified in the config subdirectory and build the
 * the name of the pseudo_master terminal.
 */
   pseudo_slave = get_cad4_terminal( cad4_nr, user_info->pw_dir );
   pseudo_master = malloc( strlen(pseudo_slave) + 1 );
   if( !pseudo_master ) {
       fprintf( stderr, "Memory Allocation Error. %s:%d Exiting.\n", 
                __FILE__, __LINE__ );
       destroy_pid_file( cd4port_file_name ); 
       exit( EXIT_FAILURE );
   }
   strcpy( pseudo_master, pseudo_slave );
   pseudo_master[5] = 'p';  /* change /dev/ttyp?  to /dev/ptyp? */

/* open() and initialize the pseudo terminals */
    fd_array[1] = open( pseudo_master, O_RDWR | O_NOCTTY | O_NONBLOCK );
    if( -1 == fd_array[1] ) {
       fprintf( stderr, "Errno = %d at line %d in file %s\n",
                errno, __LINE__, __FILE__ );
       perror( pseudo_master );
       destroy_pid_file( cd4port_file_name ); 
       exit( EXIT_FAILURE );
    }      
    if( -1 == terminal_init(fd_array[1], speed, parity) ) {
        fprintf( stderr, "Problem initializing %s. Exiting.\n",
                 pseudo_master );
        destroy_pid_file( cd4port_file_name); 
        exit( EXIT_FAILURE );
    }

/* get the range of fd's for select() right */
    if( fd_array[1] > fd_array[0] )
        s_set = fd_array[1] + 1;
    else 
        s_set = fd_array[0] + 1;

/* initialize the SYSV IPC stuff */
    key = CAD4_KEY + (key_t)strtol(cad4_nr, NULL, 10 );

    sem_id = semget( key, 1, IPC_CREAT | IPC_EXCL | MODE );
    if( -1 == sem_id ) {
        perror( "semget()" );
        close( fd_array[0] );
        close( fd_array[1] );
        destroy_pid_file( cd4port_file_name); 
        exit( EXIT_FAILURE );
    }
    ipc_ret_val = sem_set( sem_id );
    if( -1 == ipc_ret_val ) {
        sem_delete( sem_id );
        close( fd_array[0] );
        close( fd_array[1] );
        destroy_pid_file( cd4port_file_name); 
        exit( EXIT_FAILURE );
    }
    shm_id = shmget( key, sizeof(COMM_Packet), IPC_CREAT | IPC_EXCL | MODE );
    if( -1 == shm_id ) {
        sem_delete( sem_id );
        perror( "shmget()" );
        close( fd_array[0] );
        close( fd_array[1] );
        destroy_pid_file( cd4port_file_name); 
        exit( EXIT_FAILURE );
    }
    
    shared_mem = shmat( shm_id, 0, 0 );
    if( ((COMM_Packet *)-1) == shared_mem ) {
        perror( "shmat()" );
        close( fd_array[0] );
        close( fd_array[1] );
        destroy_pid_file( cd4port_file_name); 
        exit( EXIT_FAILURE );
    }

/* inform the user we are ready to go */
    fprintf( stdout, "CAD4 communication device %s.\n", 
             pseudo_slave );
    fprintf(stdout, "Pseudoterminal %s\nOpened and initialized. File descriptor %d\n", 
            pseudo_master, fd_array[1] );
    fprintf( stdout, "Shared memory attached at %p\n", shared_mem );
    fprintf( stdout, "You may now start the cad4 software.\n\n" );

/* intialize the file descriptors to be looked at with select() to a fdset 
 * on each interation of the while() loop.  Thanks to Andrew Gierth from 
 * comp.unix.programmer for pointing this out.
 */
    while( 1 ) {
        FD_ZERO( &read_set );
        FD_SET( fd_array[0], &read_set );
        FD_SET( fd_array[1], &read_set );


again:  if( (ret_val = select(s_set, &read_set, NULL, NULL, NULL)) < 0 ) {
            if( EINTR == errno ) goto again;
            perror( "select" );
        }

        for( i = 0; i < 2; i++ ) {
         if( FD_ISSET(fd_array[i], &read_set) ) {
           switch( i ) {
              case 0:        /* data at serial port, ready to read */
                 ipc_ret_val = sem_attempt_lock( sem_id );
                 n_seq++;
#ifdef TX_INCMP_PKT
                 init_packet( packet_ptr );
                 packet_ptr->n_seq = n_seq;
                 packet_ptr->direction = '>';
#else
                 init_packet( &packet );
                 packet.n_seq = n_seq;
                 packet.direction = '>';
#endif

#ifdef TX_INCMP_PKT                  
                 ser_ret_val = serial_comm( fd_array[0],
                                            fd_array[1], packet_ptr );

                 if( 0 == packet_ptr->io_bytes ) goto finish_up;
                 
                 /* copy data into shared memory region */
                 if( 0 == ser_ret_val ) {
                     memcpy( shared_mem, packet_ptr, sizeof(COMM_Packet) );
                     ipc_ret_val = sem_release( sem_id );
                 }

#else
                 packet_ptr = build_packet( fd_array[0],
                                            fd_array[1], &packet );

                 if( 0 == packet.io_bytes ) goto finish_up;

                 if( NULL != packet_ptr ) {
                     memcpy( shared_mem, &packet, sizeof(COMM_Packet) );
                     ipc_ret_val = sem_release( sem_id );

                 }
                 else {
                     ipc_ret_val = sem_release( sem_id );
                     break;
                 }
#endif
                 if( logger_attached ) {
                     ret_val = kill( pid, SIGUSR1 );
                     if( -1 == ret_val ) {
                         fprintf( stderr, "Couldn't send SIGUSR1 to PID %d: %s\n",
                                  pid, strerror(errno) );
                     }
                 }                         
                 break;

              case 1:        /* data at pty, ready to read */
                 ipc_ret_val = sem_attempt_lock( sem_id );
                 n_seq++;
#ifdef TX_INCMP_PKT
                 init_packet( packet_ptr );
                 packet_ptr->n_seq = n_seq;
                 packet_ptr->direction = '<';
#else
                 init_packet( &packet );
                 packet.n_seq = n_seq;
                 packet.direction = '<';
#endif
#ifdef TX_INCMP_PKT
                 ser_ret_val = serial_comm( fd_array[1],
                                            fd_array[0], packet_ptr );

                 if( 0 == packet_ptr->io_bytes ) goto finish_up;

                 /* copy data into shared memory region */
                 if( 0 == ser_ret_val ) {
                     memcpy( shared_mem, packet_ptr, sizeof(COMM_Packet) );
                     ipc_ret_val = sem_release( sem_id );
                 }
#else
                 packet_ptr = build_packet( fd_array[1],
                                            fd_array[0], &packet );

                 if( 0 == packet.io_bytes ) goto finish_up;

                 if( NULL != packet_ptr ) {
                     memcpy( shared_mem, &packet, sizeof(COMM_Packet) );
                     ipc_ret_val = sem_release( sem_id );

                 }
                 else {
                     ipc_ret_val = sem_release( sem_id );
                    break;
                 }
#endif
                 if( logger_attached ) {
                     ret_val = kill( pid, SIGUSR1 );
                     if( -1 == ret_val ) {
                         fprintf( stderr, "Couldn't send SIGUSR1 to PID %d: %s\n",
                                  pid, strerror(errno) );
                     }
                 }                         
                 break;

           }     /* end switch() */
         }       /* end if() */
        }        /* end for() */
    }            /* end while() */

    finish_up:
    fputs( "Closing file desciptors...\n", stdout );
    for( i = 0; i < 2; i++ )
         close( fd_array[i] );

    /* kill the cd4logger task if it is still around */
    if( -1 != kill(pid, 0) )
         kill( pid, SIGTERM );

    fputs( "Removing semaphore...\n", stdout );
    ipc_ret_val = sem_delete( sem_id );
    if( -1 == ipc_ret_val ) 
         fputs( "Removing semaphore failed.\n", stderr );
    

    fputs( "Detaching and deleting shared memory...\n", stdout );
    ipc_ret_val = shmdt( (char *)shared_mem );
    if( -1 == ipc_ret_val ) 
         perror( "shmdt()" );

    ipc_ret_val = shmctl( shm_id, IPC_RMID, 0 );
    if( -1 == ipc_ret_val ) {
         perror( "shmctl()" );
         fputs( "Use ipcs and ipcrm to check shared memory problem.\n", 
                stderr );
    }
   
    destroy_pid_file( cd4port_file_name ); 
    destroy_pid_file( pid_file_name ); 
    
    free( serial_device );
    free( pseudo_master );

    fputs( "Done.\n", stdout );
    exit( EXIT_SUCCESS );
} /*end of main() */
