#include <stdio.h>
#include <string.h>
#include <limits.h>
#include "cad4io.h"
#include "cd4port.h"

#define PKT_PRINT_FMT "%u: serial port %c cad4_run: %d bytes transferred.\n" 

#define CAD4_BUFFER_PRINT_FMT "header_byte:\t%#x\n" \
                              "data_length:\t%hd bytes\n" \
                              "sequence number:\t%d\n" \
                              "Header byte information:\n"


/* function initializes (zeros) a packet  */
void init_packet( COMM_Packet *p )
{
    memset(p->buffer, '\0', IO_DATA_LEN + IO_HEADER_LEN );
    p->n_seq = 0;
    p->io_bytes = 0;
    p->direction = INIT_DIRECTION;

    return;
}

void print_cad4_buffer( FILE *out, CAD4_Buffer *buf )
{

    char          *err_ptr, 
                  *fun_ptr;
    int           i, j,
                  extra_data = 0;

    unsigned char ch;

    ch = get_sequence_number( buf );

    fprintf( out, CAD4_BUFFER_PRINT_FMT, buf->header_byte, 
                                         buf->length.data_len, 
                                         ch );

    ch = get_error_flag( buf );
    switch( ch ) {
      case E_SUC:
         err_ptr = "Success";
         break;
      case E_TOS:
         err_ptr = "Timeout with data";
         break;
      case E_TOL:
         err_ptr = "Timeout with no data";
         break;
      case E_SEQ:
         err_ptr = "Unexpected sequence number";
         break;
      case E_CRC:
         err_ptr = "CRC error";
         break;
      case E_TYP:
         err_ptr = "Unexpected Function code";  
         break;
      case E_OVF:
         err_ptr = "Buffer overflow";
         break;
      case E_PND:
         err_ptr = "A system service fail has occured";
         break;
      default:
         err_ptr = "Unrecognized error flag.";
         break;
    }

    ch = get_function_code( buf );
    switch( ch ) {
      case FN_INIT:
         fun_ptr = "Bootstrap 11/02";
         break;
      case FN_XFR_ASC:
         fun_ptr = "Transfer ASCII buffer to 11/02";
         break;
      case FN_XFR_MEM:
         fun_ptr = "Transfer code block to 11/02";
         break;
      case FN_TR_SWR:
         extra_data = FN_TR_SWR;
         fun_ptr = "Transmit and receive switch register";
         break;
      case FN_TR_GON:
         fun_ptr = "Transmit and receive goniometer data";
         extra_data = FN_TR_GON;
         break;
      case FN_TR_ASC:
         fun_ptr = "Transmit and receive ASCII buffer";
         break;
      case FN_REQ_MEM:
         fun_ptr = "Request code block from host";
         break;
      case FN_REQ_ASC:
         fun_ptr = "Request ASCII buffer from host";
         break;
      default:
         fun_ptr = "Unrecognized function code";
         break;
    }

    fprintf( out, "Error Flag: %s\nFunction code: %#hx -- %s\n", err_ptr,
                  ch, fun_ptr );
    switch( extra_data ) {
      case FN_TR_SWR:
           fprintf( out, "Switch Register: %ho\n", buf->data.s[C4H_SWREG] );
         break;
      case FN_TR_GON:
         fputs( "\n", out );
         fprintf( out, "Switch reg:\t%ho\nRoute Flg:\t%#hx\n"
                       "Error Flg:\t%#hx\nInt Flg:\t%#hx\n", 
                        buf->data.s[C4H_SWREG],
                        buf->data.s[C4H_ROUTFL],
                        buf->data.s[C4H_ERRFL],
                        buf->data.s[C4H_INTFL] );
         break;
    }
    if( FN_XFR_MEM != ch ) {
       fputs( "Data Buffer Contents:\n", out );
       fputs( "Byte\tHex\tBit Pattern\n", out );
       for( i = 0; i < buf->length.data_len + IO_CRC_LEN; i++ ) {
            fprintf( out, "[%d]\t%#x\t",i, buf->data.c[i] );
            for( j = CHAR_BIT - 1; j >= 0; j-- ) {
                 if( buf->data.c[i] & (1 << j) )
                     fprintf( out, "%d", 1 );
                 else
                     fprintf( out, "%d", 0 );
            }
            fputs( "\n", out );
       }
    }

    fputs( "\n", out );
    return;
}

/* this function takes a CAD4_packet structure and prints the
 * contents to the specified FILE stream pointer.  Set 'verbose' variable
 * to non zero if you want to see the contents of the packets.
 */
void print_cad4_packet( FILE *out, CommList *pack, int verbose )
{

    CommList    *current, 
                *next;
    CAD4_Buffer mesg;
    size_t      n_bytes;  /* total bytes for a given packet */
    unsigned char ch;

    current = pack;

    while( NULL != current->next ) {
      next = assemble_cad4_buffer( current, &mesg, &n_bytes );
      fprintf( out, PKT_PRINT_FMT, current->pkt.n_seq,
                                   current->pkt.direction,
                                   n_bytes );
      switch( verbose ) {
       default:  /* fall through  to no printing */
        case 0:
           break;
        case 1:
           print_cad4_buffer( out, &mesg );
           break;
        case 2: {
           struct cad4_data_block blk;  /* local variable for this block */
           ch = get_function_code( &mesg );
           if( (FN_INIT == ch) ||
               (FN_TR_SWR == ch) ||
               (FN_TR_GON == ch) ) {
               fill_cad4_data_block( &mesg, &blk );
               print_cad4_data_block( out, &blk );
           }
        }
           break;
        case 3: {
           struct cad4_data_block blk;
           ch = get_function_code( &mesg );
           if( (FN_INIT == ch) ||
               (FN_TR_SWR == ch) ||
               (FN_TR_GON == ch) ) {
               fill_cad4_data_block( &mesg, &blk );
               print_cad4_data_block( out, &blk );
           }
           print_cad4_buffer( out, &mesg );
        }
           break;
      }

      current = next;  /* increment the list */
    }

    return;  
}


/* assemble_cad4_buffer() -- takes fragments of messages (i.e. packets) in
 * the form of a linked list and assembles them into a complete message
 * (i.e. a buffer).  The function returns pointer the next packet containing
 * a new message.  The calling program needs to ensure that the *p_list
 * points to the *first* packet of a message.
 */

CommList *assemble_cad4_buffer( CommList *p_list, CAD4_Buffer *mesg, size_t *bytes )
{

    char          tx_direction;  /* transmission direction, '>' or '<'  */
    size_t        total_mesg_bytes = 0,
                  io_bytes;
    CommList      *current;
    unsigned char *cur_mesg_pos;
    unsigned char tmp_buffer[BUFFER_SIZE] = {'\0'};
  
    

    current = p_list;
    tx_direction = current->pkt.direction;  /* tag the direction of this
                                             * data transfer
                                             */

/* first assemble all the data in a temporary buffer */
    cur_mesg_pos = tmp_buffer;
    while( tx_direction == current->pkt.direction ) {
       io_bytes = current->pkt.io_bytes;
       total_mesg_bytes += io_bytes;
       memcpy( cur_mesg_pos, current->pkt.buffer, io_bytes );
       cur_mesg_pos += io_bytes;
       current = current->next;
    }

/* then fill the CAD4_Buffer mesg struct */
    mesg->header_byte = tmp_buffer[HEADER_BYTE];
    memcpy( &(mesg->length.data_len), &tmp_buffer[LOW_BYTE],
            sizeof(unsigned short) );
    memcpy( mesg->data.c, &tmp_buffer[START_DATA],
            mesg->length.data_len + IO_CRC_LEN );
 
    *bytes = total_mesg_bytes;
    return current;
}

