/* the serial_comm()  function does the low level read and writes to/from the
 * serial ports and pseudoterminals.  If there is an I/O error, 
 * an error message is emitted, but the communication continues until
 * the cad4 software dies.  If the caller does not want to access what is being
 * passed between the terminal ports set the COMM_Packet pointer to NULL.
 * Otherwise, set the COMM_Packet argument to a valid address for that
 * data type.
 *
 * the build_packet() function is necessary because the interface sends
 * the return buffer back in chunks rather than doing one big write. We
 * use static variables to keep track of where we are in building the 
 * buffer.  The number of bytes in at the normal position i.e. bytes
 * [1] and [2] of the first read.  From that one can tell how much data
 * to expect (exclusive of header_byte + size + crc).  The size value + 5
 * will be the number of bytes transferred from the interface to the host
 * for a complete message.
 */

#define _XOPEN_SOURCE
#include <sys/types.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

#include "cad4io.h"
#include "cd4port.h"

int serial_comm( int rd_fd, int wr_fd, COMM_Packet *packet )
{
    size_t io_bytes;
    ssize_t r_bytes, 
            w_bytes;
    unsigned char buffer[BUFFER_SIZE];
    char *failed_dev;


    memset( buffer, '\0', BUFFER_SIZE );

/* get the number of bytes to read, then read them. */   
    if( 0 > ioctl(rd_fd, FIONREAD, (char *)&io_bytes) ) {
        failed_dev = ttyname( rd_fd );
        fprintf( stderr, "%s -- ioctl(): %s\n", failed_dev, strerror(errno) );
    }
    
    r_bytes = read( rd_fd, buffer, io_bytes );
    if( (-1 == r_bytes) || (r_bytes != io_bytes) ) {
        failed_dev = ttyname( rd_fd );
        if( -1 == r_bytes )
          fprintf( stderr,
                   "%s -- read(): %s\n", failed_dev, strerror(errno) );
        else
           fprintf( stderr,
           "%s failed on read(), errno not set.\n", failed_dev );
    }

/* now write the data to the other file descriptor */
    w_bytes = write( wr_fd, buffer, (size_t)r_bytes );
    if( (-1 == w_bytes) || (w_bytes != r_bytes) ) {
        failed_dev = ttyname( wr_fd );
        if( -1 == w_bytes )
            fprintf( stderr,
                     "%s -- write(): %s\n", failed_dev, strerror(errno) );
        else
           fprintf( stderr,
           "%s failed on write(), errno not set.\n", failed_dev );
    }
    if( -1 == tcdrain(wr_fd) ) {
        failed_dev = ttyname( wr_fd );
        fprintf( stderr,
                 "%s -- tcdrain(): %s\n", failed_dev, strerror(errno) );
    }


/* if caller wants a packet, build it up now */
    if( NULL != packet ) {
        memcpy( packet->buffer, buffer, io_bytes );
        packet->io_bytes = io_bytes;
    }

    return 0;  /* ok */
}
