#include <stdio.h>
#include <stdlib.h>

/*
 * You can use unsigned short here, but it may be more efficient to
 * grab 16-bit "hardware words" and convert them to unsigned 16-or-more-bit
 * values internally.  For now we go with the "possibly more efficient"
 * method.
 */
typedef unsigned int hw_word;	/* "word" as received from external hardware */

/*
 * Convert "stepper motor" value from hardware encoding to a floating
 * point value suitable for display.
 *
 * The w0 word contains the low 16 bits, the w1 word contains the
 * high 15 bits (as its low 15 bits) and a sign bit (as its bit 15).
 */
double stepper_value_from_hw(hw_word w1, hw_word w0) {
	/* conversion factor from "hardware encoded value" to human-readable */
	#define	STDGR	(128.0 * 4096.0 / 360.0)
#if 1
	long v31;		/* 31-bit value from hardware */
	double v32;		/* 32-bit value, including sign */

	if (w1 & 0x8000) {
		v31 = (((~w1 & 0x7fff) << 16) | (~w0 & 0xffff));
		/*
		 * The -1.0 here is because negative values are twos
		 * complement, so that (ffff,ffff) represents -1, not -0.
		 */
		v32 = -v31 - 1.0;
	} else {
		v31 = ((w1 & 0x7fff) << 16) | w0;
		v32 = v31;
	}
	return v32 / STDGR;
	#undef	STDGR
#else
	long v32;	/* must be exactly 32 bits */

	v32 = (w1 << 16) | w0;
	return v32 / STDGR;
#endif
}

/*
 * For testing, read a string in hex and make sure it is a valid "hw_word",
 * storing its value in *rval.  Returns 0 if okay, 1 if input is malformed.
 */
int hw_from_hex(const char *str, hw_word *rval) {
	char *ep;
	unsigned long t;

	t = strtol(str, &ep, 16);
	if (t > 0xffff || *ep)
		return 1;
	*rval = t;
	return 0;
}

int main(int argc, char **argv) {
	char *progname, *t;
	hw_word w0, w1;
	int i;

	if ((progname = argv[0]) == NULL)
		progname = "try";
	if (argc < 2 || (argc % 2) != 1) {
		fprintf(stderr, "usage: %s highword lowword ...\n", progname);
		exit(EXIT_FAILURE);
	}
	for (i = 1; i < argc; i += 2) {
		if (hw_from_hex(t = argv[i], &w1) ||
		    hw_from_hex(t = argv[i + 1], &w0)) {
			fprintf(stderr, "%s: malformed number `%s'\n",
			    progname, t);
			exit(EXIT_FAILURE);
		}
		printf("0x%04x, 0x%04x => %f\n", w1, w0,
		    stepper_value_from_hw(w1, w0));
	}
	exit(EXIT_SUCCESS);
}
